# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)


def chunks(listing, chunk_size):
    """
    Yield successive chunks from listing. Chunkify!
    """
    for i in range(0, len(listing), chunk_size):
        yield listing[i : i + chunk_size]



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import re

from cloud_select.logger import logger

from ..base import Instance, InstanceGroup


class AmazonInstance(Instance):
    @property
    def name(self):
        return self.data.get("InstanceType")

    def attr_price(self):
        """
        Price of an instance, USD per hour.

        This will only be populated - a single "Price" if prices are
        enabled and present for the region.
        """
        return self.data.get("Price")

    def attr_region(self):
        """
        Return the single region
        """
        return self.data.get("Region")

    def attr_description(self):
        """
        Put together a description that looks like Google's

        "208 vCPUs, 5.75 TB RAM"
        """
        # TODO - we probably went to convert GB into TB when necessary
        # This is what Google does in the description
        return f"{self.attr_cpus()} vCPUs, {self.attr_memory()} GB RAM"

    def attr_cpus(self):
        """
        Number of cpus the instance has.
        """
        return self.data.get("VCpuInfo", {}).get("DefaultVCpus")

    def attr_memory(self):
        """
        Memory is in GB
        """
        mb = self.data.get("MemoryInfo", {}).get("SizeInMiB")
        if mb:
            return int(mb / 1024)

    def attr_free_tier(self):
        """
        Determine if an instance is free tier.
        """
        return self.data.get("FreeTierEligible")

    def attr_ipv6(self):
        """
        Determine if an instance can support ipv6
        """
        return self.data.get("NetworkInfo", {}).get("Ipv6Supported")

    def attr_gpu(self):
        """
        Determine if an instance can support gpu
        """
        return self.data.get("GpuInfo") and len(self.data["GpuInfo"].get("Gpus")) > 0

    def attr_gpus(self):
        """
        Determine number of gpus an instance can support.
        """
        gpus = self.data.get("GpuInfo", {}).get("Gpus")
        if not gpus:
            return
        count = 0
        for gpu_spec in gpus:
            count += gpu_spec["Count"]
        return count


class AmazonInstanceGroup(InstanceGroup):
    """
    A Google Cloud instance group.

    An instance group stores raw data, and allows for query or
    other interaction over instances.
    """

    name_attribute = "InstanceType"
    Instance = AmazonInstance

    def filter_region(self, region):
        """
        A request to filter down to a specific region regular expression.

        The solver cannot handle this.
        """
        self.data = [x for x in self.data if re.search(region, " ".join(x["Regions"]))]

    def iter_instances(self):
        """
        Amazon provides regions together, so here we yield instance data with just one region.
        """
        for item in self.data:
            regions = item.get("Regions")
            if not regions:
                yield self.Instance(item)
                continue

            # Provide finally as single region and price
            for region in regions:
                item["Region"] = region
                item["Price"] = item.get("Prices", {}).get(region)
                yield self.Instance(item)

    def add_instance_prices(self, prices):
        """
        Add pricing information to instances

        We currently are filtering to:
         - Linux
         - OnDemand
         - USD per hour
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/pricing.html
        """
        lookup = {}
        for price in prices.data:
            if (
                "instanceType" not in price["product"]["attributes"]
                or "OnDemand" not in price["terms"]
                or price["product"]["productFamily"] != "Compute Instance"
            ):
                continue

            # I think we want BoxUsage https://docs.aws.amazon.com/cost-management/latest/userguide/ce-filtering.html
            # Note that I also see HostBoxUsage and MIA1-BoxUsage but am filtering to those that start with BoxUsage for consistency
            if price["product"]["attributes"][
                "operatingSystem"
            ] != "Linux" or not price["product"]["attributes"]["usagetype"].startswith(
                "BoxUsage"
            ):
                continue

            # The operation attribute looks like:
            # operation: ['RunInstances:0004', 'RunInstances:0200', 'RunInstances:0100', 'RunInstances']
            # And we have totally identical instances except for that, but in rates the numbers correspond to "extras" like enterprise, sql, web.
            # So for now we just want to filter to vanilla "RunInstances" which looks like the raw EC2 price
            if price["product"]["attributes"]["operation"] != "RunInstances":
                continue

            # How we will organize result entries
            location = price["product"]["attributes"]["regionCode"]
            instance_type = price["product"]["attributes"]["instanceType"]

            assert len(price["terms"]["OnDemand"]) == 1
            priceid = list(price["terms"]["OnDemand"].keys())[0]
            for _, rate in price["terms"]["OnDemand"][priceid][
                "priceDimensions"
            ].items():
                if rate["unit"].lower() == "hrs":
                    if instance_type not in lookup:
                        lookup[instance_type] = {}
                    if location in lookup[instance_type]:
                        logger.warning(
                            f"Found two rates for {instance_type} in {location} - we will choose one but this likely should not happen."
                        )
                    # These are provided as strings, since the original data is completely string
                    lookup[instance_type][location] = float(rate["pricePerUnit"]["USD"])

        # Add prices to instances we have prices for
        for instance in self.data:

            # Set to unreasonably high so it's not a choice
            if instance["InstanceType"] in lookup:

                # Make a list of prices that matches regions
                region_prices = {}
                for region in instance["Regions"]:
                    if region in lookup[instance["InstanceType"]]:
                        region_prices[region] = lookup[instance["InstanceType"]][region]
                    instance["Prices"] = region_prices



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import os
import shutil
import time

import cloud_select.utils as utils
from cloud_select.logger import logger


class Cache:
    """
    Cache instances, prices, and other things.
    """

    def __init__(self, cache_dir, cache_expire=128):
        self.disable_cache = False
        if cache_expire == 0:
            self.disable_cache = True
        self.cache_dir = cache_dir
        self._cache = {}
        self._cache_expire_hours = cache_expire

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[cloud-select-cache]"

    def clear(self, force=False):
        """
        Clear the cache (with confirmation).
        """
        if not force and not utils.confirm_action(
            "Are you sure you want to clear the cache? "
        ):
            return
        if os.path.exists(self.cache_dir):
            shutil.rmtree(self.cache_dir)

    def set(self, cloud_name, items, datatype, cls=None):
        """
        Given a result, cache if the user has cache enabled.

        Allow to set a custom data encoder for cloud data.
        """
        if self.disable_cache:
            return

        cache_file = self.get_cache_name(cloud_name, datatype)
        cache_dir = os.path.dirname(cache_file)

        # Ensure cache directory exists
        utils.mkdir_p(cache_dir)

        # Don't write empty data
        if not items:
            logger.warning(
                "No data found for {cloud_name} {datatype}, not writing {cache_file}."
            )
            return

        # Save to "memory" and filesystem cache
        if cloud_name not in self._cache:
            self._cache[cloud_name] = {}
        self._cache[cloud_name][datatype] = items
        utils.write_json(items, cache_file, cls=cls)
        logger.debug("{cloud_name} {datatype} written to {cache_file}.")

    def get_cache_name(self, cloud_name, name):
        """
        Return a json cache entry for a given cloud provider and data type
        """
        return os.path.join(self.cache_dir, cloud_name, f"{name}.json")

    def is_expired(self, cloud_name, datatype):
        """
        Determine if cache data is expired.
        """
        if self.disable_cache:
            return False
        cache_file = self.get_cache_name(cloud_name, datatype)
        stats = os.stat(cache_file)

        # Convert cache_expire hours to seconds
        expire_seconds = self._cache_expire_hours * 60 * 60

        # And determine if the time now since modified is greater than expire
        return (time.time() - stats.st_mtime) > expire_seconds

    def exists(self, cloud_name, datatype):
        """
        Determine if cache data exists.
        """
        if self.disable_cache:
            return False
        return os.path.exists(self.get_cache_name(cloud_name, datatype))

    def get(self, cloud_name, datatype):
        """
        Given a cache name (typically matching the endpoint) retrieve if exists.
        If provided and endpoint, wrap the result with the endpoint. Otherwise,
        return the json result.
        """
        # First effort - get from memory
        if cloud_name in self._cache and datatype in self._cache[cloud_name]:
            return self._cache[cloud_name][datatype]

        # Now look for filesystem
        cache_file = self.get_cache_name(cloud_name, datatype)
        if not os.path.exists(cache_file):
            return

        # Load the cache, reset if error loading (invalid json)
        try:
            return utils.read_json(cache_file)
        except Exception as e:
            logger.warning(f"Cache entry {cache_file} has corrupt json: {e}, removing.")
            os.remove(cache_file)



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)


import re

import cloud_select.defaults as defaults
import cloud_select.main.cache as cache
import cloud_select.main.cloud as clouds
import cloud_select.main.schemas as schemas
import cloud_select.main.solve as solve
from cloud_select.logger import logger

from .settings import Settings


class Client:
    """
    Cloud select client.
    """

    def __init__(self, **kwargs):
        validate = kwargs.get("validate", True)
        self.quiet = kwargs.get("quiet", False)
        self.settings = Settings(kwargs.get("settings_file"), validate)
        self.set_clouds(kwargs.get("clouds"))

        # Set the cache and expiration
        self.cache = cache.Cache(
            kwargs.get("cache_dir") or self.settings.cache_dir or defaults.cache_dir,
            cache_expire=kwargs.get("cache_expire")
            or self.settings.cache_expire
            or defaults.cache_expire,
        )

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[cloud-select-client]"

    def set_clouds(self, listing=None):
        """
        Set client cloud classes that will be attempted.
        """
        # The user can specify clouds on the command line or settings
        # command line takes preference!
        selection = list(set(self.settings.clouds if not listing else listing))

        # Reset previously init-d clouds
        self._clouds = {}
        self._cloudclass = {}
        if selection:
            for item in selection:
                item = item.lower()
                if item in clouds.cloud_names:
                    self._cloudclass[item] = clouds.clouds[item]
        else:
            self._cloudclass = clouds.clouds

    def get_clouds(self, force=False, lookup=False):
        """
        Return listing of clouds we can authenticate to, or we have data for.
        """
        # Return cached set, unless forced
        if self._clouds and not force:
            return self._clouds if lookup else list(self._clouds.values())

        # We should always be able to get cloud classes, even without auth
        # The class knows how to parse the data types into a standard space
        for cloud_name, CloudClass in self._cloudclass.items():

            # Regions default to settings then defaults
            cloud_settings = getattr(self.settings, cloud_name)
            self._clouds[cloud_name] = CloudClass(
                regions=cloud_settings.get("regions"),
                cache_only=self.settings.cache_only,
            )
        return self._clouds if lookup else list(self._clouds.values())

    def instances(self):
        """
        Return instances for clouds we have cached data for.
        """
        return self.load_cache("instances")

    def prices(self):
        """
        Return instance pricing for clouds we have cached data for.
        """
        return self.load_cache("prices")

    def load_cache(self, key):
        """
        Load a named entry from the cache known to the cloud provider.

        This function does not check cache expiration, but just
        returns data we have available.
        """
        items = {}
        for cloud in self.get_clouds():
            if self.cache.exists(cloud.name, key):
                # Load data via the cloud provider
                data = self.cache.get(cloud.name, key)
                items[cloud.name] = getattr(cloud, f"load_{key}")(data)
        return items

    def update_from_cache(self, items, datatype):
        """
        Given a data type, update from the cache.
        """
        # For every cloud class we have...
        for cloud in self.get_clouds():

            # We have the data and it's expired OR we don't have it - update it
            if cloud.name not in items or self.cache.is_expired(cloud.name, datatype):
                func = getattr(cloud, datatype, None)

                # This should not happen, but let's be careful
                if not func:
                    logger.warning(
                        "Cannot call {datatype} function for cloud {cloud_name}, function is missing."
                    )
                    continue

                # Get updated items, but don't update data if we cannot (usually not authenticated)
                updated = func()
                if not updated:
                    continue
                self.cache.set(cloud.name, updated, datatype, cls=updated.Encoder)
                items[cloud.name] = updated
        return items

    def instance_select(self, max_results=20, **kwargs):
        """
        Select an instance.
        """
        # Start with already cached data
        instances = self.update_from_cache(self.instances(), "instances")
        if not instances:
            logger.exit(
                "You don't have any clouds to search instances or cached data. Set credentials or get an offline cache."
            )

        # Only makes sense to get prices if we have instances!
        prices = {}
        if self.settings.disable_prices is not True:
            prices = self.update_from_cache(self.prices(), "prices")

        # Attributes that can't go into the solver
        region = kwargs.get("region")
        if "region" in kwargs:
            del kwargs["region"]

        # By here we have a lookup *by cloud) of instance groups
        # Filter down kwargs (properties) to those relevant to instances
        # This is not actually a solve, but it used to be and we kept the name
        properties = solve.Properties(schemas.instance_properties, **kwargs)
        solver = solve.Solver()

        # 1. write mapping of common features into functions
        # 2. filter down to desired set based on these common functions
        for cloud_name, instance_group in instances.items():

            # Do we have a request to filter by region?
            if region is not None:
                instance_group.filter_region(region)

            # Do we have prices for the cloud?
            if cloud_name in prices:
                instance_group.add_instance_prices(prices[cloud_name])

            # Generate facts for instances
            solver.add_instances(cloud_name, instance_group)

        solver.add_properties(properties.defined)

        # Select the instances!
        selected = solver.solve().get("instance") or []

        # Do we have a request for a pattern to include or exclude?
        if selected:
            for pattern in properties.include_list or []:
                selected = [x for x in selected if re.search(pattern, x)]
            for pattern in properties.exclude_list or []:
                selected = [x for x in selected if not re.search(pattern, x)]

        # Assemble back into complete data based on instance name
        return [instances[x[0]].generate_row(x[1]) for x in selected]



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import json

import cloud_select.main.table as table
import cloud_select.utils as utils
from cloud_select.main import Client


def main(args, parser, extra, subparser):

    utils.ensure_no_extra(extra)

    cli = Client(
        quiet=args.quiet,
        settings_file=args.settings_file,
        cache_dir=args.cache_dir,
        cache_expire=args.cache_expire,
        clouds=args.clouds,
    )

    # Update config settings on the fly
    cli.settings.update_params(args.config_params)

    # If max results is 0, set to None (no limit)
    if args.max_results == 0:
        args.max_results = None

    # Are we writing ASP to an output file?
    out = args.out
    delattr(args, "out")

    # And select the instance (this output is for the ASP)
    instances = cli.instance_select(**args.__dict__)

    # Print instances to a table
    if args.json:
        print(json.dumps(instances, indent=4))
    elif out:
        utils.write_json(out, instances)
    else:
        t = table.Table(instances)
        t.table(
            title="Cloud Instances Selected",
            sort_by=args.sort_by,
            limit=args.max_results,
            ascending=args.ascending,
        )



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import os

from rich.console import Console
from rich.table import Table as RichTable

import cloud_select.defaults as defaults
import cloud_select.main.colors as colors
from cloud_select.logger import logger

here = os.path.dirname(os.path.abspath(__file__))


class Table:
    """
    Format a result into a table.
    """

    def __init__(self, data):
        self.data = data
        self.max_widths = {}
        self.ensure_complete()

    def available_width(self, columns):
        """
        Calculate available width based on fields we cannot truncate (urls)
        """
        # We will determine column width based on terminal size
        try:
            width = os.get_terminal_size().columns
        except OSError:
            width = 120

        # Calculate column width
        column_width = int(width / len(columns))
        updated = width

        for _, needed in self.max_widths.items():
            updated = updated - needed

        # We don't have enough space
        if updated < 0:
            logger.warning("Terminal is too small to correctly render!")
            return column_width

        # Otherwise, recalculate column width taking into account truncation
        # We use the updated smaller width, and break it up between columns
        # that don't have a max width
        return int(updated / (len(columns) - len(self.max_widths)))

    def ensure_complete(self):
        """
        If any data missing fields, ensure they are included
        """
        if isinstance(self.data, list):
            self.ensure_complete_list()
        # We don't check other types for now

    def ensure_complete_list(self):
        """
        Given a list result, check the fields.
        """
        fields = set()
        for entry in self.data:
            [fields.add(x) for x in entry.keys()]

        # Ensure fields are present
        for entry in self.data:
            for field in fields:

                if field not in entry:
                    entry[field] = ""

    def table_columns(self):
        """
        Shared function to return consistent table columns
        """
        # Plan to remove empty columns with count 0
        column_counts = {x: 0 for x, _ in self.data[0].items()}

        # Count entries for each column
        for entry in self.data:
            for column, value in entry.items():
                if value is not None:
                    column_counts[column] += 1

        # Get column titles
        columns = []
        contenders = list(self.data[0].keys())
        for column in contenders:
            if column_counts[column] == 0:
                continue
            columns.append(column)
        return columns

    def table_rows(self, columns, limit=25):
        """
        Shared function to yield rows as a list
        """
        # All keys are lowercase
        column_width = self.available_width(columns)
        for i, row in enumerate(self.data):

            # have we gone over the limit?
            if limit and i > limit:
                return

            parsed = []
            for column in columns:
                content = str(row[column]) if row[column] is not None else ""
                if content is not None and len(content) > column_width:
                    content = content[:column_width] + "..."
                parsed.append(content)
            yield parsed

    def table(self, limit=None, title=None, sort_by=None, ascending=False):
        """
        Pretty print a table of results.
        """
        table = RichTable(title=title)

        # No dependencies!
        if not self.data:
            print("There are no results to report.")
            return

        # Get column titles and unique colors
        columns = self.table_columns()
        column_colors = colors.get_rich_colors(len(columns))

        for i, column in enumerate(columns):
            title = " ".join([x.capitalize() for x in column.split("_")])
            table.add_column(title, style=column_colors[i])

        # If we want sorting, filter down to those that have it
        if sort_by is not None and sort_by in self.data[0].keys():
            subset = [x for x in self.data if x.get(sort_by) not in [None, ""]]
            if not subset:
                logger.warning(f"Using filter {sort_by} to sort removes all results.")
                return

            # Break into two groups - first those that have the value, then we will add the rest
            self.data = sorted(subset, key=lambda x: x[sort_by], reverse=not ascending)

        # Add rows
        for row in self.table_rows(columns, limit=limit):
            table.add_row(*row)

        # And print!
        console = Console()
        console.print(table, justify="left")

        # Get the color of the price column to add description
        for descriptor, text in defaults.table_descriptors.items():
            if descriptor in columns:
                console.print(f"{text}", style=column_colors[columns.index(descriptor)])



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)


import os

import cloud_select.utils as utils

install_dir = utils.get_installdir()
reps = {"$install_dir": install_dir, "$root_dir": os.path.dirname(install_dir)}

# The default settings file in the install root
default_settings_file = os.path.join(reps["$install_dir"], "settings.yml")

# User home
userhome = os.path.expanduser("~/.cloud-select")

# The user settings file can be created to over-ride default
user_settings_file = os.path.join(userhome, "settings.yml")
cache_dir = os.path.join(userhome, "cache")
cache_expire = 128  # one week is 128 hours

# Fields and arguments for client
sort_by_fields = ["name", "memory", "cpus", "gpus", "price"]
table_descriptors = {"price": "*Price in USD/hour"}

# variables in settings that allow environment variable expansion
allowed_envars = ["HOME"]

# The default GitHub registry with recipes (for docgen)
github_url = "https://github.com/converged-computing/cloud-select"



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)


import json

from cloud_select.logger import logger


class CloudProvider:
    """
    A base class for a cloud provider
    """

    name = "cloud"

    def __init__(self):

        # If we weren't created with settings, add empty
        if not hasattr(self, "settings"):
            from cloud_select.main.settings import Settings

            self.settings = Settings()

    def __str__(self):
        return str(self.__class__.__name__)

    def load_instances(self, data):
        """
        Load instance data from json.
        """
        raise NotImplementedError

    def instances(self):
        """
        Retrieve all instances for a cloud.
        """
        raise NotImplementedError

    def fail_message(self, message):
        """
        Shared message and empty return if auth not set
        """
        logger.info(f"{self.name}: cannot retrieve {message}.")
        return []


class CloudDataEncoder(json.JSONEncoder):
    """
    Make our group and instance classes json serializable.

    Always return the data!
    """

    def default(self, o):
        return o.data


class CloudData:
    """
    An "abstract" class that stores a single data object

    This also stores the encoder for any cloud object so
    we can encode it.
    """

    Encoder = CloudDataEncoder

    def __init__(self, items):
        self.data = items
        self.lookup = {}
        self.create_lookup()

    def create_lookup(self):
        """
        If we have a name attribute and a list of data, cache a lookup for later
        """
        if not (hasattr(self, "name_attribute") and isinstance(self.data, list)):
            return
        key = self.name_attribute
        for item in self.data:
            if key not in item:
                continue
            self.lookup[item[key]] = item


class Instance(CloudData):
    """
    Base of an instance.

    This class defines the different common attributes that we want
    to expose. If a cloud instance (json) result differs, it should
    override this function.
    """

    @property
    def attribute_getters(self):
        """
        function names we can call to get an instance attribute.
        """
        fields = []
        for func in dir(self):
            if func.startswith("attr_"):
                fields.append(func)
        return fields

    # Attributes shared between clouds (maybe)
    def attr_description(self):
        return self.data.get("description")

    def attr_zone(self):
        return self.data.get("zone")

    @property
    def cloud(self):
        """
        Return the name of the cloud (module) derived from path.
        """
        # We can get the cloud name here
        module = self.__class__.__module__
        return module.replace("cloud_select.main.cloud.", "").split(".")[0]

    def generate_row(self):
        """
        Given an instance name, return a row with the cloud
        name and other attributes.
        """
        return {
            "cloud": self.cloud,
            "name": self.name,
            "memory": self.attr_memory(),
            "price": self.attr_price(),
            "cpus": self.attr_cpus(),
            "gpus": self.attr_gpus(),
            "region(s)": self.attr_region(),
            "description": self.attr_description(),
        }

    @property
    def name(self):
        return self.data.get("name")


class InstanceGroup(CloudData):
    """
    A cloud instance group.

    An instance group stores raw data, and allows for query or
    other interaction over instances.
    """

    # If we don't have an instance class, return as dict
    Instance = dict

    def filter_region(self, region):
        """
        Filter by a region (not required)
        """
        pass

    def generate_row(self, name):
        """
        Given an instance name, return a row with the cloud
        name and other attributes.
        """
        # We assume that we have the item in the lookup
        if name not in self.lookup:
            raise ValueError(f"{name} is not known to {self}")
        return self.Instance(self.lookup[name]).generate_row()

    def add_instance_prices(self, prices):
        """
        Add pricing information to instances
        """
        raise NotImplementedError

    def iter_instances(self):
        """
        Each client knows how to instantiate its instance type
        """
        for item in self.data:
            yield self.Instance(item)


class Prices(CloudData):
    """
    Cloud pricing
    """

    pass



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import re

from ..base import Instance, InstanceGroup


class GoogleCloudInstance(Instance):
    def attr_cpus(self):
        """
        Number of cpus the instance has.
        """
        return self.data.get("guestCpus")

    def attr_memory(self):
        """
        Memory is in GB
        """
        return int(self.data["memoryMb"] / 1024)

    def attr_region(self):
        """
        Return the (| joined) listing of regions
        """
        return self.data.get("zone")

    def attr_free_tier(self):
        """
        Determine if an instance is free tier.
        https://cloud.google.com/free/docs/free-cloud-features#compute

        This is a best effort ESTIMATION based on the name of the instance,
        saying it _could_ be free tier but isn't necessarily if you've used
        your monthly allowance. We will want to add that preamble somewhere.
        """
        return "micro" in self.name

    def attr_ipv6(self):
        """
        Determine if an instance can support ipv6
        As far as I can tell - they all can be configured with a subnet that does?

        https://cloud.google.com/compute/docs/ip-addresses/configure-ipv6-address
        """
        return True

    def attr_price(self):
        """
        Price of an instance, USD per hour.

        This is not added for Google yet
        """
        return self.data.get("price")

    def attr_gpu(self):
        """
        Determine if an instance can support gpu

        https://cloud.google.com/compute/docs/gpus

        To run NVIDIA A100 GPUs, you must use the accelerator-optimized (A2) machine type.
        """
        accels = self.data.get("accelerators", []) or []
        return len(accels) > 1 or "gpu" in self.name.lower()

    def attr_gpus(self):
        """
        Determine number of gpus an instance can support.

        https://cloud.google.com/compute/docs/gpus

        To run NVIDIA A100 GPUs, you must use the accelerator-optimized (A2) machine type.
        """
        accels = self.data.get("accelerators", []) or []
        count = None

        # List of [{'guestAcceleratorType': 'nvidia-tesla-a100', 'guestAcceleratorCount': 8}]
        if accels:
            for accel in accels:
                found_count = accel.get("guestAcceleratorCount")
                if found_count and count is None or found_count > count:
                    count = found_count
        return count


class GoogleCloudInstanceGroup(InstanceGroup):
    """
    A Google Cloud instance group.

    An instance group stores raw data, and allows for query or
    other interaction over instances.
    """

    name_attribute = "name"
    Instance = GoogleCloudInstance

    def filter_region(self, region):
        """
        A request to filter down to a specific region regular expression.

        The solver cannot handle this.
        """
        self.data = [x for x in self.data if re.search(region, x["zone"])]

    def add_instance_prices(self, prices):
        """
        Add pricing information to instances

        Prices have these categories:
          - {'Compute', 'License', 'Network', 'Storage'}
        And these usage types:
          - {'Commit1Mo', 'Commit1Yr', 'Commit3Yr', 'OnDemand', 'Preemptible'}

        For now we will filter to "Compute" and "OnDemand" but these could be changed.
        """
        # Filter down to compute
        data = [
            x
            for x in prices.data
            if x["category"]["resourceFamily"] == "Compute"
            and x["category"]["usageType"] == "OnDemand"
        ]
        assert data
        # TODO - here we have a lsiting with snapshot, CPU, RAM, and I suspect we will need to combine attributes to get costs
        # for instances. This is a TODO I want help with. For now, we do nothing



#!/usr/bin/env python

# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import argparse
import os
import sys

import cloud_select
import cloud_select.defaults as defaults
import cloud_select.main.cloud as cloud
import cloud_select.main.schemas as schemas
from cloud_select.logger import setup_logger


def add_instance_arguments(command):
    """
    Derive the attributes requested for an instance.
    """
    # All all options (based on type) except for command, which comes in "extra"
    for name, attrs in schemas.instance_properties.items():
        if name == "command":
            continue
        typ = attrs.get("type")

        # This is currently just envars, we will add separately as append args.
        if not typ and "oneOf" in attrs:
            continue

        choices = attrs.get("enum")
        default_type = str
        default = attrs.get("default")

        # It's either a string...
        if typ == "number" or "number" in typ:
            default_type = int
        elif typ == "boolean" or "boolean" in typ:
            default_type = bool
            default = None

        elif typ == "array":
            typ = attrs["items"]["type"]

            # Right now we have a list to select one thing from...
            if "enum" in attrs["items"]:
                choices = attrs["items"]["enum"]

        if default_type == bool:
            command.add_argument(
                f"--{name}",
                help=attrs.get("description") or f"The --{name} flag.",
                action=argparse.BooleanOptionalAction,  # THis ensures the default can be None
                default=default,
            )
        else:
            command.add_argument(
                f"--{name}",
                help=attrs.get("description") or f"The --{name} flag.",
                type=default_type,
                choices=choices,
                default=default,
            )


def get_parser():
    parser = argparse.ArgumentParser(
        description="Cloud Select",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Global Variables
    parser.add_argument(
        "--debug",
        dest="debug",
        help="use verbose logging to debug.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--quiet",
        dest="quiet",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--verbose",
        dest="verbose",
        help="print additional solver output (atoms).",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--settings-file",
        dest="settings_file",
        help="custom path to settings file.",
    )

    parser.add_argument(
        "--cache-dir",
        dest="cache_dir",
        help="directory for data cache (defaults to ~/.cloud-select/cache).",
    )

    parser.add_argument(
        "--max-results",
        dest="max_results",
        help="Maximum results to return per cloud provider.",
        type=int,
        default=25,
    )
    parser.add_argument(
        "--cloud",
        dest="clouds",
        help="one or more clouds to include (if not provided, all are attempted).",
        choices=cloud.cloud_names,
        action="append",
    )

    parser.add_argument(
        "--sort-by",
        dest="sort_by",
        help="Sort by a result attribute.",
        choices=defaults.sort_by_fields,
    )
    parser.add_argument(
        "--asc",
        dest="ascending",
        help="Sort results ascending instead of descending (default)",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--cache-expire",
        dest="cache_expire",
        help="Expire the cache (and recreate) after this many hours (defaults to 168, one week). Set to 0 to not store a cache.",
        default=128,
    )

    # On the fly updates to config params
    parser.add_argument(
        "-c",
        dest="config_params",
        help=""""customize a config value on the fly to ADD/SET/REMOVE for a command
cloud-select -c set:key:value <command> <args>
cloud-select -c add:registry:/tmp/registry <command> <args>
cloud-select -c rm:registry:/tmp/registry""",
        action="append",
    )

    parser.add_argument(
        "--version",
        dest="version",
        help="show software version.",
        default=False,
        action="store_true",
    )

    subparsers = parser.add_subparsers(
        help="cloud_select actions",
        title="actions",
        description="actions",
        dest="command",
    )

    # print version and exit
    subparsers.add_parser("version", description="show software version")

    # Local shell with client loaded
    shell = subparsers.add_parser(
        "shell",
        description="shell into a Python session with a client.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    shell.add_argument(
        "--interpreter",
        "-i",
        dest="interpreter",
        help="python interpreter",
        choices=["ipython", "python", "bpython"],
        default="ipython",
    )

    # Install a known recipe from the registry
    instance = subparsers.add_parser(
        "instance",
        description="select an instance.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    instance.add_argument(
        "--json",
        help="dump output as json to terminal",
        default=False,
        action="store_true",
    )

    instance.add_argument(
        "--out",
        dest="out",
        help="Write instances as json to output file.",
    )
    # Add attributes from spec
    add_instance_arguments(instance)
    return parser


def run():
    """
    run_cloud_select is the entrypoint to the singularity-hpc client.
    """

    parser = get_parser()

    def help(return_code=0):
        """print help, including the software version and active client
        and exit with return code.
        """

        version = cloud_select.__version__

        print("\nSingularity Registry (HPC) Client v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    if args.debug is True:
        os.environ["MESSAGELEVEL"] = "DEBUG"

    # Show the version and exit
    if args.command == "version" or args.version:
        print(cloud_select.__version__)
        sys.exit(0)

    setup_logger(
        quiet=args.quiet,
        debug=args.debug,
    )

    # retrieve subparser (with help) from parser
    helper = None
    subparsers_actions = [
        action
        for action in parser._actions
        if isinstance(action, argparse._SubParsersAction)
    ]
    for subparsers_action in subparsers_actions:
        for choice, subparser in subparsers_action.choices.items():
            if choice == args.command:
                helper = subparser
                break

    # Does the user want a shell?
    if args.command == "instance":
        from .instance import main
    elif args.command == "config":
        from .config import main
    elif args.command == "shell":
        from .shell import main

    # Pass on to the correct parser
    return_code = 0
    try:
        main(args=args, parser=parser, extra=extra, subparser=helper)
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1

    help(return_code)


if __name__ == "__main__":
    run()



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import json
import re

import cloud_select.utils as utils
from cloud_select.logger import logger

from ..base import CloudProvider
from .instance import AmazonInstanceGroup
from .prices import AmazonPrices


class AmazonCloud(CloudProvider):
    """
    An Amazon cloud provider wrapper
    """

    name = "aws"

    def __init__(self, **kwargs):
        self.regions = kwargs.get("regions") or ["us-east-1"]
        self.project = None
        self.ec2_client = None
        self.pricing_cli = None

        # This currently has two pieces - billing and instances (different APIs)
        self._set_services(kwargs.get("cache_only", False))
        super(AmazonCloud, self).__init__()

    def prices(self):
        """
        Use the API to retrieve and return prices to cache.
        """
        if not self.has_pricing_auth:
            return self.fail_message("prices, authentication not set.")

        # Get services first - there are almost 2k! Look for compute engine
        logger.info(f"Retrieving prices for {self.name}.")

        # Keep a price that matches any region we care about
        regex = "(%s)" % "|".join(self.regions)

        next_token = ""
        prices = []
        while True:
            response = self.pricing_cli.get_products(
                ServiceCode="AmazonEC2", NextToken=next_token
            )
            if not response.get("NextToken"):
                break
            next_token = response.get("NextToken")
            # The prices are actually string - so let's search region of interest via regex
            for pricestr in response["PriceList"]:
                if not re.search(regex, pricestr):
                    continue
                prices.append(json.loads(pricestr))
            print(f"{len(prices)} total aws prices matching {regex}...", end="\r")
        print()

        return self.load_prices(response)

    def instances(self):
        """
        Use the API to retrieve (and return) instances within a set of regions.

        We do a little extra work to add the region attribute to make it
        accessible for filtering.
        """
        if not self.has_instance_auth:
            return self.fail_message("instances, authentication not set.")

        logger.info(f"Retrieving instances for {self.name}")

        # Start with a lookup so we can attach regions
        machine_types = []
        lookup = {}

        # Get instance types by region so we have stored in metadata
        for region in self.regions:
            response = self.ec2_client.describe_instance_type_offerings(
                DryRun=False,
                LocationType="region",
                Filters=[{"Name": "location", "Values": self.regions}],
            )
            if not response["InstanceTypeOfferings"]:
                logger.warning(
                    f"No instance types found for region selection {region} - are you sure it is correct?"
                )

            # Organize by machine type for now
            for x in response["InstanceTypeOfferings"]:
                machine_types.append(x["InstanceType"])
                if x["InstanceType"] not in lookup:
                    lookup[x["InstanceType"]] = []
                lookup[x["InstanceType"]].append(region)

        # Make sure we have unique machine types
        machine_types = list(set(machine_types))

        # From what I can tell, we don't have next pages for this smaller list.
        # We can only ask in increments of 100
        instances = []
        for chunk in utils.chunks(machine_types, 100):
            response = self.ec2_client.describe_instance_types(
                DryRun=False, InstanceTypes=chunk
            )
            for new_instance in response.get("InstanceTypes", []):
                new_instance["Regions"] = lookup.get(new_instance["InstanceType"])
                instances.append(new_instance)

        # Return a wrapped set of instances
        return self.load_instances(instances)

    def load_prices(self, data):
        """
        Load prices data from json instal class
        """
        return AmazonPrices(data)

    def load_instances(self, data):
        """
        Load instance data from json.
        """
        return AmazonInstanceGroup(data)

    def _set_services(self, cache_only=False):
        """
        Connect to needed amazon clients.

        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_instance_types
        """
        # Cut out early if only cache is wanted.
        if cache_only is True:
            self.has_instance_auth = False
            self.has_pricing_auth = False
            return

        import boto3

        self.ec2_client = boto3.client("ec2")
        self.pricing_cli = boto3.client("pricing")

        try:
            self.ec2_client.describe_instances()
            self.has_instance_auth = True
        except Exception as e:
            logger.warning(f"Unable to authenticate to Amazon Web Services EC2: {e}")
            self.has_instance_auth = False

        # Note: purposefully set to false because we don't have an API token to test
        try:
            self.pricing_cli.describe_services(ServiceCode="AmazonEC2")
            self.has_pricing_auth = True
        except Exception as e:
            logger.warning(f"Unable to authenticate to Amazon Web Services EC2: {e}")
            self.has_pricing_auth = False



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)


import sqlite3
from functools import partial, update_wrapper

from cloud_select.logger import logger


class with_connection:
    """
    Provide the function with a connection with open and close.

    This is used as a decorator. When added, self.conn of the
    class is instantiated for interacting with, and cleaned up
    at the end.
    """

    def __init__(self, func):
        update_wrapper(self, func)
        self.func = func

    def __get__(self, obj, objtype):
        return partial(self.__call__, obj)

    def __call__(self, cls, *args, **kwargs):
        cls.conn = cls.db.cursor()
        res = self.func(cls, *args, **kwargs)
        cls.conn.close()
        cls.conn = None
        return res


class Database:
    """
    In-memory database to query instances.
    """

    def __init__(self, filename=":memory:"):
        self.db = sqlite3.connect(filename)
        self.conn = None
        self.execute(create_instances_table)
        self.filename = filename
        # Lookup of table id to instance class
        # If too much memory, can write database to temporary file
        self.lookup = {}

    @with_connection
    def execute(self, command, fetchall=False):
        """
        Create the database, creating the instances table.
        """
        try:
            if fetchall:
                return self.conn.execute(command).fetchall()
            return self.conn.execute(command)
        except Exception as e:
            logger.exit(e)

    @with_connection
    def execute_many(self, table, items):
        """
        Create the database, creating the instances table.
        """
        if not items:
            return

        # Keys should be consistent across
        keys = list(items[0].keys())
        values = [tuple(item[x] for x in keys) for item in items]
        insert_values = ",".join(keys)
        insert_qs = ",".join("?" * len(keys))
        try:
            self.conn.executemany(
                f"insert into {table}({insert_values}) values ({insert_qs})", values
            )
        except Exception as e:
            logger.exit(e)

    def filter_instances(self, props):
        """
        Use properties to filter instances down to a desired set based.
        """
        basequery = "SELECT DISTINCT cloud, instance FROM instances"
        # No properties,
        if not props:
            return self.conn.execute(f"{basequery};").fetchall()

        query = ""

        # Assemble rest of query
        for _, key in enumerate(props):
            if not query:
                query = basequery
            else:
                query += f"INTERSECT {basequery}"

            # Case 1: we have a range with lookup min/max
            value = props[key]

            # This returns a completed sql statement
            if key.startswith("range:"):
                value = parse_range(key, value)
            else:
                value = parse_value(value)
            query += f" WHERE {value}\n"

        logger.debug(query)
        return {"instance": self.execute(f"{query};", fetchall=True)}

    def add_instance(self, cloud_name, instance):
        """
        Generic ability to add instance to flat database
        """
        # We will do a bulk insert
        rows = []

        for func in instance.attribute_getters:
            value = getattr(instance, func)()
            attr = func.replace("attr_", "")
            # If we don't have data, we don't add it
            if value in [None, ""]:
                continue

            # Have instance key be name and region
            item = {"cloud": cloud_name, "instance": instance.name, "attribute": attr}

            # Switch off and structure based on what kind of attribute we have.
            if isinstance(value, str):
                item.update({"value": value, "value_bool": None, "value_number": None})
            elif isinstance(value, bool):
                item.update(
                    {"value": None, "value_bool": int(value), "value_number": None}
                )
            else:
                item.update({"value": None, "value_bool": None, "value_number": value})
            rows.append(item)

        self.execute_many("instances", rows)


# The table will be flat - either we have a value (text) or number but not both
# for each attribute
create_instances_table = """
CREATE TABLE IF NOT EXISTS instances (
    id integer PRIMARY KEY,
    cloud text NOT NULL,
    instance text NOT NULL,
    attribute text NOT NULL,
    value text NULLABLE,
    value_bool number NULLABLE,
    value_number number NULLABLE
);"""

# Example of foreign key
# 	FOREIGN KEY (project_id) REFERENCES projects (id)


def parse_value(key, value):
    """
    Parse a provided attribute value into an query.
    If we get here, it's not a range.

    True/False -> 1, 0
    string     -> "string"
    """
    if isinstance(value, bool):
        return (
            f"value_bool IS NOT NULL AND attribute='{key}' and value_bool={int(value)}"
        )
    if isinstance(value, str):
        return f"value IS NOT NULL AND attribute='{key}' and value='{value}'"
    if isinstance(value, (int, float)):
        return f"value_number IS NOT NULL AND attribute='{key}' and value_number={int(value)}"

    # We should not get here
    raise ValueError("A value that isn't bool or string should not be parsed.")


def parse_range(key, value):
    """
    Given a stated range, e.g.,:

    range_gpus: {'min': 2, 'max': None}

    Return either a min/max rule, or just a min,or just a max.
    """
    # If we get to a range without min/max this is an error (and shouldn't happen)
    key = key.replace("range:", "", 1)
    min_value = value["min"]
    max_value = value["max"]

    # All of these are numbers, so we only care about value_number
    if min_value is not None and max_value is not None:
        return f"value_number IS NOT NULL AND attribute='{key}' AND value_number >= {min_value} AND value_number <= {max_value}"

    if min_value is not None:
        return f"value_number IS NOT NULL AND attribute='{key}' and value_number >= {min_value}"

    # Max value set
    return f"value_number IS NOT NULL AND attribute='{key}' and value_number <= {max_value}"



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import re
import time

from cloud_select.logger import logger

from ..base import CloudProvider
from .instance import GoogleCloudInstanceGroup
from .prices import GoogleCloudPrices


class GoogleCloud(CloudProvider):
    """
    A Google Cloud provider wrapper

    We allow the client init to proceed given authentication is not
    possible as it can provide data served from a cache, wrapping
    available instances.
    """

    name = "google"

    def __init__(self, **kwargs):
        self.regions = kwargs.get("regions") or ["us-east1", "us-west1", "us-central1"]
        self.project = None
        self.compute_cli = None
        self.billing_cli = None
        try:
            self._set_services(kwargs.get("cache_only"))
        except Exception as e:
            logger.warning(f"Unable to authenticate to Google Cloud: {e}")
            self.has_auth = False
        super(GoogleCloud, self).__init__()

    def prices(self):
        """
        Use the API to retrieve and return prices to cache.
        """
        if not self.has_auth:
            return self.no_auth_message("prices, authentication not set.")

        # Get services first - there are almost 2k! Look for compute engine
        logger.info(
            f"Retrieving prices for {self.name} - this might take a few minutes."
        )
        services = self._retry_request(self.billing_cli.services().list())
        services = [
            x for x in services["services"] if x["displayName"] == "Compute Engine"
        ]

        # Bail out if we don't have one service
        if len(services) != 1:
            return self.no_auth_message("prices, cannot find Compute Engine service.")

        # We want to add "global" to our regions to search
        regions = self.regions + ["global"]

        # Get metadata about compute (ALL prices here) and handle nextPage tokens
        # https://cloud.google.com/billing/docs/reference/rest/v1/services.skus/list
        # There are millions of prices - we will filter down to compute engine
        # instances, but this first call still takes a long time. I think this
        # could be hugely helped if we are able to provide a cache of filtered prices.
        # The default currency returned is dollars (USD).
        prices = []
        page_token = None
        while True:
            items = self._retry_request(
                self.billing_cli.services()
                .skus()
                .list(parent=services[0]["name"], pageToken=page_token)
            )

            # We can only reasonably save entries from our service regions of interest
            prices += [
                sku
                for sku in items.get("skus", [])
                if any(x in sku["serviceRegions"] for x in regions)
            ]
            print(f"{len(prices)} total results...", end="\r")
            if not items.get("nextPageToken"):
                break
            page_token = items["nextPageToken"]

        # Handle pagination
        return self.load_prices(prices)

    def instances(self):
        """
        Use the API to retrieve (and return) instances within a set of regions.
        """
        if not self.has_auth:
            return self.no_auth_message("instances, authentication not set.")

        logger.info(f"Retrieving instances for {self.name}")

        # Regular expression to determine if zone in region
        regexp = "^(%s)" % "|".join(self.regions)

        # Retrieve zones, filter down to selected regions
        zones = self._retry_request(self.compute_cli.zones().list(project=self.project))
        zones = [z for z in zones["items"] if re.search(regexp, z["name"])]

        # Retrieve machine types available in zones.
        # https://cloud.google.com/compute/docs/regions-zones/
        machine_types = []
        for zone in zones:
            request = self.compute_cli.machineTypes().list(
                project=self.project, zone=zone["name"]
            )
            machine_types += self._retry_request(request)["items"]

        # Get accelerator types (for GPU) to add to them?
        # TODO - I don't see where we can get memory for GPUs :(
        # https://cloud.google.com/compute/docs/reference/rest/v1/acceleratorTypes
        # accels = self._retry_request(self.compute_cli.acceleratorTypes().aggregatedList(project=self.project))

        # Return a wrapped set of instances
        return self.load_instances(machine_types)

    def load_prices(self, data):
        """
        Load prices data from json instal class
        """
        return GoogleCloudPrices(data)

    def load_instances(self, data):
        """
        Load instance data from json.
        """
        return GoogleCloudInstanceGroup(data)

    def _set_services(self, cache_only=False):
        """
        Use Google Discovery Build to generate an API client for compute and billing.
        """
        if cache_only is True:
            self.has_auth = False
            return
        import google.auth
        import google_auth_httplib2
        import googleapiclient
        import httplib2
        from googleapiclient.discovery import build as discovery_build

        # Get default credentials. If there is an exception, caught by init function
        # google.auth.DefaultCredentialsError
        creds, project = google.auth.default(
            scopes=["https://www.googleapis.com/auth/cloud-platform"]
        )
        self.project = project

        def build_request(_, *args, **kwargs):
            """
            See https://googleapis.github.io/google-api-python-client/docs/thread_safety.html

            Note that the first positional arg (http) is required despite not being used here.
            """
            new_http = google_auth_httplib2.AuthorizedHttp(creds, http=httplib2.Http())
            return googleapiclient.http.HttpRequest(new_http, *args, **kwargs)

        # Discovery client for Google Cloud Compute
        # https://cloud.google.com/compute/docs/reference/rest/v1/instances
        authorized_http = google_auth_httplib2.AuthorizedHttp(creds)
        self.compute_cli = discovery_build(
            "compute",
            "v1",
            cache_discovery=False,
            http=authorized_http,
            requestBuilder=build_request,
        )
        self.billing_cli = discovery_build(
            "cloudbilling",
            "v1",
            cache_discovery=False,
            http=authorized_http,
            requestBuilder=build_request,
        )
        self.has_auth = True

    def _retry_request(self, request, timeout=2, attempts=3):
        """
        The Google Python API client frequently has BrokenPipe errors. This
        function takes a request, and executes it up to number of retry,
        each time with a 2* increase in timeout.
        """
        import googleapiclient

        try:
            return request.execute()
        except BrokenPipeError as e:
            if attempts > 0:
                time.sleep(timeout)
                return self._retry_request(
                    request, timeout=timeout * 2, attempts=attempts - 1
                )
            raise e
        except googleapiclient.errors.HttpError as e:
            if attempts > 0:
                time.sleep(timeout)
                return self._retry_request(
                    request, timeout=timeout * 2, attempts=attempts - 1
                )
            raise e
        except Exception as e:
            if attempts > 0:
                time.sleep(timeout)
                return self._retry_request(
                    request, timeout=timeout * 2, attempts=attempts - 1
                )
            raise e



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

from ..base import Prices


class GoogleCloudPrices(Prices):
    """
    Google Compute Engine prices
    """

    pass



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import re

from cloud_select.logger import logger


class Properties:
    """
    Properties from a schema to go into a solve.
    """

    # Do not go into the solve (but if defined, used before or after)
    not_solver_properties = ["include_list", "exclude_list"]

    # These aren't added/developed yet (patterns for initial list with "-")
    skip_patterns = ["price-per-hour"]

    # expected to have <name>, <name>_min, <name>_max
    range_properties = ["cpus", "instance_storage", "memory", "gpu_memory"]

    def __init__(self, properties, **kwargs):
        self._allowed = properties
        self.set_properties(**kwargs)

    def set_properties(self, **kwargs):
        """
        Given an unknown set of command line arguments, parse into known
        instance properties. This resets properties / defined on the instance.

        We assume that properties coming from argumnts might have a "-" instead of _.
        """
        skip_pattern = "(%s)" % "|".join(self.skip_patterns)
        properties = {
            k: v
            for k, v in kwargs.items()
            if k.replace("_", "-") in self._allowed and not re.search(skip_pattern, k)
        }
        # Instance level properties (to be deleted from selection, used separately)
        self.set_not_solver_properties(properties)

        # The user is asking for these defined
        self.defined = {k: v for k, v in properties.items() if v is not None}

        # GPU is handled specially since there is a boolean and min/max/number
        self.set_gpu_properties()

        # Here we need custom handling of min / max ranges. E.g., if a single value is
        # provided, we look exactly for that. If a min or max is provided, we look for
        # the range.
        self.set_range_properties()

    def set_range_properties(self):
        """
        Consolidate range into either min/max or one value.
        """
        for prop in self.range_properties:
            self._set_range_properties(prop)

    def _set_range_properties(self, prop):
        """
        Set range properties for one property in defined.
        """
        # Case 1: explicitly asking for a number - honor it.
        range_props = [f"{prop}_min", f"{prop}_max"]
        if prop in self.defined:
            for range_prop in range_props:
                if range_prop in self.defined:
                    del self.defined[range_prop]
            return

        # Case 2: Min and/or max is set - we have a range
        values = {"min": None, "max": None}
        for range_prop in range_props:
            if range_prop in self.defined:
                value = self.defined[range_prop]

                # Sanity check - always >= 0
                if value and value < 0:
                    logger.exit(f"Found {range_prop} value < 0: {value}.")

                values[range_prop.replace(f"{prop}_", "")] = value
                del self.defined[range_prop]

        # If we have min and max, assert max >= min
        if values["min"] is not None and values["max"] is not None:
            if values["max"] < values["min"]:
                logger.exit(
                    f"Found max {values['max']} less than a min {values['min']}, invalid."
                )

        # Only add if we've found a min and/or max
        if values["min"] is not None or values["max"] is not None:
            self.defined[f"range:{prop}"] = values

    def set_gpu_properties(self):
        """
        Parse desired properties into single GPU request.
        There is both a min/max/number specification and a general flag.
        """
        # Properties that ask for a specific count
        gpu_props = ["gpus", "gpus_max", "gpus_min"]

        # Case 1: they just provided the "I want gpu don't care the details" flag
        if "gpu" in self.defined and not any(x for x in gpu_props if x in self.defined):
            return

        # Case 2: they provided the gpu flag but have other specs, delete GPU flag
        if "gpu" in self.defined:
            del self.defined["gpu"]

        # Case 3: Just a gpus is set (exact number) delete the rest
        self._set_range_properties("gpus")

    def set_not_solver_properties(self, properties):
        """
        These properties are used later (e.g., to filter a final set based on name).
        """
        # Instance level properties (to be deleted from selection, used separately)
        for prop in self.not_solver_properties:
            if prop not in properties:
                setattr(self, prop, None)
                continue
            setattr(self, prop, properties.get(prop))
            del properties[prop]



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

## ContainerConfig Schema

import cloud_select.defaults as defaults

schema_url = "http://json-schema.org/draft-07/schema"

# This is also for latest, and a list of tags

# The simplest form of aliases is key/value pairs
keyvals = {
    "type": "object",
    "patternProperties": {
        "\\w[\\w-]*": {"type": "string"},
    },
}

# instance_properties
# This is a shared list of attributes that could be defined (and shared across clouds)
# TODO could we do some mapping of zones to generic regions (e.g., east/central/west)

# Things that need common mapping:
# - architecture
# - cpus/cores (ranges)?

# We need (in clingo facts) ways to compare things

instance_properties = {
    "gpus": {
        "type": "number",
        "description": "Number of gpus needed (assumes gpu-enabled wanted) and sets gpu min and max to the same value.",
    },
    "gpus-min": {
        "type": "number",
        "description": "Min of total number of GPUs (e.g., 4) if --gpus-max not set, is infinity",
    },
    "gpus-max": {
        "type": "number",
        "description": "Max of total number of GPUs (e.g., 4) if --gpus-min not set, it is 0",
    },
    "gpu": {"type": "boolean", "description": "GPU-enabled instance type."},
    "gpu-memory": {
        "type": "number",
        "description": "Total memory across GPUs in GiB (e.g., 4) sets --gpu-memory-min and --gpu-memory-max to the same value",
    },
    "gpu-memory-min": {
        "type": "number",
        "description": "Min of total memory across GPUs in GiB (e.g., 4 GiB) if --gpu-memory-max not set, is infinity",
    },
    "gpu-memory-max": {
        "type": "number",
        "description": "Max of total memory across GPUs in GiB (e.g., 4 GiB) if --gpu-memory-min not set, it is 0",
    },
    "gpu-model": {"type": "string", "description": "GPU model name."},
    "region": {
        "type": "string",
        "description": "Regular expression or string to search for in region name.",
    },
    "hypervisor": {
        "type": "string",
        "description": "Hypervisor.",
        "enum": ["xen", "nitro"],
    },
    "memory": {
        "type": "number",
        "description": "Total memory needed and sets memory min and max to the same value.",
    },
    "memory-min": {
        "type": "number",
        "description": "Min memory needed in GiB. if max not set, is infinity",
    },
    "memory-max": {
        "type": "number",
        "description": "Max memory needed in GiB if min not set, it is 0",
    },
    "instance-storage": {
        "type": "number",
        "description": "Amount of local instance storage in GiB. Sets --instance-storage-min and -max to the same value.",
    },
    "instance-storage-min": {
        "type": "number",
        "description": "Minimum amount of local instance storage in GiB. If --instance-storage-max not set, is infinity.",
    },
    "instance-storage-max": {
        "type": "number",
        "description": "Maximum amount of local instance storage in GiB. If --instance-storage-min not set, is 0.",
    },
    "cpus": {
        "type": "number",
        "description": "Number of vcpus available to the instance type. Sets min and -max to the same value.",
    },
    "cpus-min": {
        "type": "number",
        "description": "Minimum number of vcpus available to the instance type. If max not set, is infinity.",
    },
    "cpus-max": {
        "type": "number",
        "description": "Maximum number of vcpus available to the instance type. If min not set, is 0.",
    },
    "price-per-hour": {
        "type": "number",
        "description": "Price/hour in dollars (e.g., 0.09) (sets min and max to the same value)",
    },
    "price-per-hour-min": {
        "type": "number",
        "description": "Minimum price/hour in dollars. If max not set, is infinity.",
    },
    "price-per-hour-max": {
        "type": "number",
        "description": "Maximum price/hour in dollars. If min not set, is 0.",
    },
    "free-tier": {"type": "boolean", "description": "Free tier only."},
    "ipv6": {"type": "boolean", "description": "Instance Types that support IPv6"},
    "cpu-arch": {
        "type": "string",
        "description": "architecture of the CPU",
        "enum": ["x86_64/amd64", "x86_64_mac", "i386", "arm64"],
    },
    "cpu-vendor": {
        "type": "string",
        "description": "manufacturer of CPU",
        "enum": ["amd", "intel", "aws"],
    },
    "gpu-vendor": {"type": "string", "enum": ["nvidia"]},
    "disk-type": {
        "type": "array",
        "description": "type of disk, hard disk or solid state",
        "items": {"type": "string", "enum": ["hdd", "ssd"]},
    },
    "exclude-list": {
        "type": "array",
        "items": {"type": "string"},
        "description": "instance types which should be excluded w/ regex syntax (Example: m[1-2]\.*)",  # noqa
    },
    "include-list": {
        "type": "array",
        "items": {"type": "string"},
        "description": "instance types which should be included w/ regex syntax (Example: m[1-2]\.*)",  # noqa
    },
}

#      --network-encryption                             Instance Types that support automatic network encryption in-transit
#      --network-interfaces int                         Number of network interfaces (ENIs) that can be attached to the instance (sets --network-interfaces-min and -max to the same value)
#      --network-interfaces-max int                     Maximum Number of network interfaces (ENIs) that can be attached to the instance If --network-interfaces-min is not specified, the lower bound will be 0
#      --network-interfaces-min int                     Minimum Number of network interfaces (ENIs) that can be attached to the instance If --network-interfaces-max is not specified, the upper bound will be infinity
#      --network-performance int                        Bandwidth in Gib/s of network performance (Example: 100) (sets --network-performance-min and -max to the same value)
#      --network-performance-max int                    Maximum Bandwidth in Gib/s of network performance (Example: 100) If --network-performance-min is not specified, the lower bound will be 0
#      --network-performance-min int                    Minimum Bandwidth in Gib/s of network performance (Example: 100) If --network-performance-max is not specified, the upper bound will be infinity
#      --nvme                                           EBS or local instance storage where NVME is supported or required
#      --placement-group-strategy string                Placement group strategy: [cluster, partition, spread]
#  -u, --usage-class string                             Usage class: [spot or on-demand]
#      --virtualization-type string                     Virtualization Type supported: [hvm or pv]
#  -o, --output string           Specify the output format (table, table-wide, one-line, interactive)
#      --profile string          AWS CLI profile to use for credentials and config
#  -r, --region string           AWS Region to use for API requests (NOTE: if not passed in, uses AWS SDK default precedence)
#      --sort-by string          Specify the field to sort by. Quantity flags present in this CLI (memory, gpus, etc.) or a JSON path to the appropriate instance type field (Ex: ".MemoryInfo.SizeInMiB") is acceptable. (default ".InstanceType")
#      --sort-direction string   Specify the direction to sort in (ascending, asc, descending, desc) (default "ascending")

## Settings.yml (loads as json)

cloud_properties = {"regions": {"type": "array", "items": {"type": "string"}}}

# Currently all of these are required
settings_properties = {
    "cache_dir": {"type": "string"},
    "config_editor": {"type": "string"},
    "cache_only": {"type": "boolean"},
    "aws": {
        "type": "object",
        "properties": cloud_properties,
        "additionalProperties": False,
        "required": ["regions"],
    },
    "google": {"type": "object", "properties": cloud_properties},
    "disable_prices": {
        "type": "boolean",
        "description": "Do not add prices as variable.",
        "default": False,
    },
    "sort-by": {
        "type": "string",
        "description": "Sort by an attribute of interest.",
        "enum": defaults.sort_by_fields,
    },
    "max-results": {
        "type": "number",
        "description": "Maximum results to return per cloud provider.",
        "default": 20,
    },
    "instances": {
        "type": "object",
        "properties": instance_properties,
        "additionalProperties": False,
    },
    "clouds": {"type": "array", "items": {"type": "string", "enum": ["google", "aws"]}},
}

settings = {
    "$schema": schema_url,
    "title": "Settings Schema",
    "type": "object",
    "required": [
        "clouds",
        "google",
        "aws",
    ],
    "properties": settings_properties,
    "additionalProperties": False,
}



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)


from .database import Database


class Solver:
    """
    Class to orchestrate a selection Solver.

    This originally used clingo, but since clingo doesn't support floats (yikes)
    I refactored to create an in memory database to query.
    """

    def __init__(self):
        """
        Create a solver to generate facts and rules for an ASP program.
        """
        self.db = Database()
        self.data = {}

    def add_properties(self, props):
        """
        Add properties to the setup
        """
        # These are properties we will query for
        self.data["properties"] = props

    def add_instances(self, cloud_name, instance_group):
        """
        Add cloud instances to the database
        """
        for instance in instance_group.iter_instances():
            self.db.add_instance(cloud_name, instance)

    def solve(self):
        """
        A manual solve just uses python filtering!
        """
        return self.db.filter_instances(self.data["properties"])
